from .binary.morgan import MorganBinaryClassifier
from .binary.mordred import MordredBinaryClassifier
from .binary.classic import ClassicBinaryClassifier
from .binary.rdkit import RdkitBinaryClassifier
from .binary.eosembedding import ErsiliaBinaryClassifier
from .regression.morgan import MorganRegressor
from .regression.eosembedding import ErsiliaRegressor